/**********************************************************
 **
 ** jal_ctrl.c : parser for JAL control functions
 **
 ** Copyright (c) 2004-2005, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include "../libcore/pf_cmd.h"
#include "../libcore/pf_expr.h"
#include "jal_expr.h"
#include "jal_blck.h"
#include "jal_file.h"
#include "jal_ctrl.h"

/*
 * NAME
 *   jal_parse_for
 *
 * DESCRIPTION
 *   implement FOR
 *
 * PARAMETERS
 *   pf       : 
 *
 * RETURN
 *   none
 *
 * NOTES
 *   FOR expr LOOP ... END LOOP
 * this becomes:
 *   tmp = expr
 *   label_top:
 *     if (!tmp) goto label_end
 *     block
 *     tmp--
 *     goto label_top
 *   label_end:
 */
void jal_parse_for(pfile_t *pf)
{
  static const char *ewords[] = {"end"};
  size_t    ew;
  value_t   expr;
  boolean_t codegen_disable;
  label_t   lbl_top;
  label_t   lbl_end;
  value_t   val;
  value_t   zero;

  expr = jal_parse_expr(pf);

  if (pf_token_is(pf, pf_token_current, pfile_log_none, "using")) {
    pf_token_get(pf, pf_token_next);
    val = VALUE_NONE;
    pf_token_to_value(pf, pfile_log_err, &val);
    if (VALUE_NONE != val) {
      pf_token_get(pf, pf_token_next);
    }
    if (value_is_const(expr)) {
      /* check that expr will fit in the resulting variable */
      if (value_is_bit(val)) {
        if (((1UL << value_sz_get(val)) - 1) < value_const_get(expr)) {
          pfile_log(pf, pfile_log_err, "expr doesn't fit into variable");
        } else if (value_is_boolean(val)) {
          pfile_log(pf, pfile_log_warn, 
              "BIT values shouldn't be used in loops");
        }
      } else {
        if (value_sz_get(expr) > value_sz_get(val)) {
          pfile_log(pf, pfile_log_err, "expr doesn't fit into variable");
        }
      }
    }
  } else {
    val = jal_file_loop_var_get(pf, expr);
  }

  codegen_disable = (value_is_const(expr) && !value_const_get(expr));

  if (codegen_disable) {
    pfile_codegen_disable(pf, boolean_true);
  }

  zero = pfile_constant_get(pf, 0, VARIABLE_DEF_NONE);
  pfile_cmd_op_add(pf, operator_assign, &val, zero, VALUE_NONE);
  value_release(zero);

  if (pf_token_is(pf, pf_token_current, pfile_log_err, "loop")) {
    pf_token_get(pf, pf_token_next);
  }

  lbl_top = pfile_label_alloc(pf, 0);
  lbl_end = pfile_label_alloc(pf, 0);

  if (lbl_end) {
    pfile_cmd_branch_add(pf, cmd_branchtype_goto,
      cmd_branchcond_none, lbl_end, VALUE_NONE, 0, 0);
  }

  pfile_cmd_label_add(pf, lbl_top);
  jal_block_process(pf, JAL_BLOCK_PROCESS_FLAG_NONE,
      COUNT(ewords), ewords, &ew);
  pfile_cmd_op_add(pf, operator_incr, &val, VALUE_NONE, 
      VALUE_NONE);
  if (lbl_end) {
    value_t tmp;

    pfile_cmd_label_add(pf, lbl_end);
    tmp = VALUE_NONE;
    pfile_cmd_op_add(pf, operator_lt, &tmp, val, expr);
    pfile_cmd_branch_add(pf, cmd_branchtype_goto, 
        cmd_branchcond_true, lbl_top, tmp, 0, 0);
    value_release(tmp);
  }
  if (pf_token_is(pf, pf_token_current, pfile_log_err, "loop")) {
    pf_token_get(pf, pf_token_next);
  }
  label_release(lbl_end);
  label_release(lbl_top);
  if (codegen_disable) {
    pfile_codegen_disable(pf, boolean_false);
  }
  value_release(val);
  value_release(expr);
}

/*
 * NAME
 *   jal_parse_forever
 *
 * DESCRIPTION
 *   implement FOREVER
 *
 * PARAMETERS
 *   pf: 
 *
 * RETURN
 *   none
 *
 * NOTES
 *   FOREVER LOOP ... END LOOP
 * this becomes:
 *   label_top:
 *      block
 *      goto label_top
 */
void jal_parse_forever(pfile_t *pf)
{
  static const char *ewords[] = {"end"};
  size_t  ew;
  label_t lbl_top;

  if (pf_token_is(pf, pf_token_current, pfile_log_err, "loop")) {
    pf_token_get(pf, pf_token_next);
  }

  lbl_top = pfile_label_alloc(pf, 0);

  pfile_cmd_label_add(pf, lbl_top);
  jal_block_process(pf, JAL_BLOCK_PROCESS_FLAG_NONE,
      COUNT(ewords), ewords, &ew);
  if (pf_token_is(pf, pf_token_current, pfile_log_err, "loop")) {
    pf_token_get(pf, pf_token_next);
  }
  if (lbl_top) {
    pfile_cmd_branch_add(pf, cmd_branchtype_goto,
        cmd_branchcond_none, lbl_top, VALUE_NONE, 0, 0);
  }
  label_release(lbl_top);
}

/*
 * NAME
 *   jal_parse_while
 *
 * DESCRIPTION
 *   implement WHILE
 *
 * PARAMETERS
 *   pf: 
 *
 * RETURN
 *   none
 *
 * NOTES
 *   WHILE bexpr LOOP ... END LOOP
 * this becomes:
 *   tmp = bexpr
 *   label_top:
 *     if (!tmp) goto label_end
 *     block
 *     goto label_top
 *   label_end: 
 */
void jal_parse_while(pfile_t *pf)
{
  static const char *ewords[] = {"end"};
  size_t    ew;
  label_t   lbl_top;
  label_t   lbl_end;
  value_t   expr;
  boolean_t codegen_disable;

  lbl_top = pfile_label_alloc(pf, 0);
  lbl_end = pfile_label_alloc(pf, 0);
  expr    = VALUE_NONE;

  pfile_cmd_label_add(pf, lbl_top);
  expr = jal_parse_expr(pf);

  codegen_disable = (value_is_const(expr) && !value_const_get(expr));
  if (codegen_disable) {
    pfile_codegen_disable(pf, boolean_true);
  }
  if (lbl_end && expr) {
    pfile_cmd_branch_add(pf, cmd_branchtype_goto,
        cmd_branchcond_false, lbl_end, expr, 0, 0);
  }
  value_release(expr);
  if (pf_token_is(pf, pf_token_current, pfile_log_err, "loop")) {
    pf_token_get(pf, pf_token_next);
  }
  jal_block_process(pf, JAL_BLOCK_PROCESS_FLAG_NONE,
      COUNT(ewords), ewords, &ew);
  if (pf_token_is(pf, pf_token_current, pfile_log_err, "loop")) {
    pf_token_get(pf, pf_token_next);
  }
  pfile_cmd_branch_add(pf, cmd_branchtype_goto,
      cmd_branchcond_none, lbl_top, VALUE_NONE, 0, 0);
  pfile_cmd_label_add(pf, lbl_end);
  if (codegen_disable) {
    pfile_codegen_disable(pf, boolean_false);
  }
  label_release(lbl_end);
  label_release(lbl_top);
}

/*
 * NAME
 *   jal_skip_if
 *
 * DESCRIPTION
 *   skip an if block
 *
 * PARAMETERS
 *   pf       : pfile
 *   eword_ct : # of end words
 *   ewords   : end words
 *   ew       : [out] holds the terminating word
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this is a rather wacky result of the way the IF statement works in
 *   JAL. JAL, having no preprocessor, uses `IF cexpr THEN...' like the
 *   preprocess #if ...
 */
static void jal_skip_if(pfile_t *pf, size_t eword_ct,
    const char **ewords, size_t *ew)
{
  size_t ii;

  do {
    /* embedded IF */
    while (pf_token_is(pf, pf_token_current, pfile_log_none, "if")) {
      pf_token_get(pf, pf_token_next);
      do {
        static const char *end[] = { "end" };
        jal_skip_if(pf, 1, end, &ii);
      } while (!pf_token_is_eof(pf) 
          && !pf_token_is(pf, pf_token_current, pfile_log_none, "if"));
      pf_token_get(pf, pf_token_next);
    }
    for (ii = 0; 
        (ii < eword_ct) && !pf_token_is(pf, pf_token_current, pfile_log_none,
                                        ewords[ii]); 
        ii++)
      ;
    if (pf_token_is(pf, pf_token_current, pfile_log_none, "end")) {
      if (!pf_token_is(pf, pf_token_next, pfile_log_none, "if")) {
        ii = eword_ct;
      }
    } else {
      pf_token_get(pf, pf_token_next);
    }
  } while ((ii == eword_ct) && !pf_token_is_eof(pf));
  if (ii < eword_ct) {
    *ew = ii;
  }
}

/*
 * NAME
 *   jal_parse_if
 *
 * DESCRIPTION
 *   implement IF
 *
 * PARAMETERS
 *   pf: 
 *
 * RETURN
 *   none
 *
 * NOTES
 *   IF expr THEN .. [ELSIF expr THEN ...] [ELSE ...] END IF
 * becomes:
 *   release for each ELSIF:
 *     if !expr goto label_next
 *       block
 *       goto label_end
 *     label_next
 *     ..
 *     label_end
 */
void jal_parse_if(pfile_t *pf)
{
  static const char *ewords[] = { "end", "elsif", "else" };
  enum {
    JAL_PARSE_IF_END,
    JAL_PARSE_IF_ELSIF,
    JAL_PARSE_IF_ELSE
  };
  size_t    ew;
  label_t   lbl_end;
  boolean_t else_skip;

  lbl_end = pfile_label_alloc(pf, 0);

  /* else_skip is set to TRUE if any of the preceding IF clauses
     has evaluated to a constant TRUE which means the ELSE can
     never be hit */
  else_skip = boolean_false;
  do {
    value_t   expr;
    boolean_t skip_clause;
    label_t   lbl_next;

    expr = jal_parse_expr(pf);
    skip_clause = boolean_false;
    if (value_is_const(expr)) {
      if (pfile_flag_test(pf, PFILE_FLAG_SHOWDIRECTIVES)) {
        pfile_log(pf, pfile_log_warn, "compiler directive");
      }
      if (value_const_get(expr)) {
        else_skip = boolean_true;
      } else {
        skip_clause = boolean_true;
      }
    }
    if (skip_clause) {
      pfile_codegen_disable(pf, boolean_true);
      lbl_next = LABEL_NONE;
    } else {
      lbl_next = pfile_label_alloc(pf, 0);
      if (lbl_next) {
        pfile_cmd_branch_add(pf, cmd_branchtype_goto,
            cmd_branchcond_false, lbl_next, expr, 0, 0);
      }
    }
    value_release(expr);
    if (pf_token_is(pf, pf_token_current, pfile_log_err, "then")) {
      pf_token_get(pf, pf_token_next);
    }
    if (skip_clause) {
      jal_skip_if(pf, COUNT(ewords), ewords, &ew);
    } else {
      jal_block_process(pf, JAL_BLOCK_PROCESS_FLAG_NONE,
          COUNT(ewords), ewords, &ew);
    }
    if (lbl_end) {
      pfile_cmd_branch_add(pf, cmd_branchtype_goto,
          cmd_branchcond_none, lbl_end, VALUE_NONE, 0, 0);
    }
    if (skip_clause) {
      pfile_codegen_disable(pf, boolean_false);
    } else {
      pfile_cmd_label_add(pf, lbl_next);
      label_release(lbl_next);
    }
  } while ((JAL_PARSE_IF_ELSIF == ew) && !pf_token_is_eof(pf));
  if (JAL_PARSE_IF_ELSE == ew) {
    if (else_skip) {
      jal_skip_if(pf, COUNT(ewords), ewords, &ew);
    } else {
      jal_block_process(pf, JAL_BLOCK_PROCESS_FLAG_NONE,
          1, ewords, &ew);
    }
  }
  if (pf_token_is(pf, pf_token_current, pfile_log_err, "if")) {
    pf_token_get(pf, pf_token_next);
  }
  pfile_cmd_label_add(pf, lbl_end);
  label_release(lbl_end);
}

